--/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   FPGA Design 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/06/15 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* Naming convention:  http://dz.ee.ethz.ch/en/information/hdl-help/vhdl-naming-conventions.html
--*
--***************************************************************************************************************
--*
--* This is the analyzers main entity. It hooks up all the cores for synchronizing, sampling and triggering.
--*
--***************************************************************************************************************
----------------------------------------------------------------------------------
-- core.vhd
--
-- Copyright (C) 2006  Michael Poppitz
-- 
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.
--
-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
--
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc.,
-- 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
--
----------------------------------------------------------------------------------
--
-- Details: http://www.sump.org/projects/analyzer/
--
-- The core contains all "platform independent" modules and provides a
-- simple interface to those components. The core makes the analyzer
-- memory type and computer interface independent.
--
-- This module also provides a better target for test benches as commands can
-- be sent to the core easily.
--
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

library UNISIM;
use UNISIM.VComponents.all;

entity core is
  generic(
    CHANNELS : integer
    );
  port (
    ClkxCI               : in  std_logic;  -- sampling clock
    ClkxCBI              : in  std_logic;  -- 180 degree phase shifted sampling clk
    resetxRI             : in  std_logic;
    cmdDatxDI            : in  std_logic_vector (31 downto 0);  --command to control the La core
    opcodexDI            : in  std_logic_vector(7 downto 0);  --opcode, determines what the meaning of cmd is
    executexSI           : in  std_logic;  --high for one clock cycle ==> new cmd/opcode available
    channelsxDI          : in  std_logic_vector (31 downto 0);  --signal input
    flagExtClkxSO        : out std_logic;  -- signals which source should be use to clock the core
    sampleClkxSO         : out std_logic;  -- outputs the sample clock if the divider is set 1 or more
    memoryClkxCO         : out std_logic;  -- data output clock signal
    memoryOut            : out std_logic_vector (31 downto 0);
    memoryWrite          : out std_logic;
    sample_finish_IRQxSO : out std_logic;  --interrupt-request-flag for the AVR. (finished sampling)
    sample_clrflagxSI    : in  std_logic);  --input to clear the interrupt flag
end core;

architecture Behavioral of core is

  component decoder
    port (
      ClkxCI             : in  std_logic;
      opcodexDI          : in  std_logic_vector (7 downto 0);
      dataxDI            : in  std_logic_vector(31 downto 0);
      executexSI         : in  std_logic;
      wrTrigUnitSelxSO   : out std_logic;
      wrTrigMaskxSO      : out std_logic;
      wrTrigValxSO       : out std_logic;
      wrTrigFlagsxSO     : out std_logic;
      wrTrigDelayxSO     : out std_logic;
      wrSpeedxSO         : out std_logic;
      wrFwdxSO           : out std_logic;
      wrBwdxSO           : out std_logic;
      flagDemuxxSO       : out std_logic;
      flagFilterxSO      : out std_logic;
      flagExtClkxSO      : out std_logic;
      flagFallingEdgexSO : out std_logic;
      runxSO             : out std_logic;
      resetCorexRO       : out std_logic);
  end component;

  component sync
    generic (
      CHANNELS : integer);
    port (
      ClkxCI      : in  std_logic;
      ClkxCIB     : in  std_logic;
      ChannelsxDI : in  std_logic_vector (CHANNELS-1 downto 0);
      FilterxS    : in  std_logic;
      DemuxxS     : in  std_logic;
      FallingxS   : in  std_logic;
      SyncedxDO   : out std_logic_vector (CHANNELS-1 downto 0));
  end component;

  component sampler
    generic (
      CLOCK_DIV_WIDTH : integer);
    port (
      ClkxCI         : in  std_logic;
      RstxRI         : in  std_logic;
      RunxSI         : in  std_logic;
      samplexDI      : in  std_logic_vector (31 downto 0);
      cfgDataxDI     : in  std_logic_vector (CLOCK_DIV_WIDTH-1 downto 0);
      wrDividerxSI   : in  std_logic;
      samplexDO      : out std_logic_vector (31 downto 0);
      sampleValidxSO : out std_logic;
      sampleRatexSO  : out std_logic);
  end component;

  component trigger
    generic (
      DELAY_LEN : integer;
      UNITS_LEN : integer);
    port (
      ClkxCI           : in  std_logic;
      RstxRI           : in  std_logic;
      DatxDI           : in  std_logic_vector (31 downto 0);
      DatValidxSI      : in  std_logic;
      DemuxedxSI       : in  std_logic;
      WriteUnitxSI     : in  std_logic;
      CfgDatxDI        : in  std_logic_vector (31 downto 0);
      WriteMaskxSI     : in  std_logic;
      WriteMatchValxSI : in  std_logic;
      WriteFlagsxSI    : in  std_logic;
      WriteDelayxSI    : in  std_logic;
      ArmxSI           : in  std_logic;
      RunxSO           : out std_logic;
      DatxDO           : out std_logic_vector(31 downto 0);
      DatValidxSO      : out std_logic);
  end component;

  component sampleCounter
    generic (
      CNT_WIDTH : integer);
    port (
      ClkxCI         : in  std_logic;
      RstxRI         : in  std_logic;
      samplexDI      : in  std_logic_vector(31 downto 0);
      sampleValidxSI : in  std_logic;
      runxSI         : in  std_logic;
      wrTopxSI       : in  std_logic;
      cfgDataxDI     : in  std_logic_vector(CNT_WIDTH-1 downto 0);
      donexSO        : out std_logic;
      samplexDO      : out std_logic_vector(31 downto 0);
      sampleValidxSO : out std_logic);
  end component;

  constant TRIG_UNITS_LEN : integer := 3;  -- number of bits, i.e. 3 give 8 trigger units

  constant FWD_BWD_WIDTH : integer := 25;  -- width of forward and backward counters in bits

  signal syncedInputxDN, syncedInputxDP : std_logic_vector (31 downto 0) := (others => '0');
  --signal sampleClkxC, sampleClkxCB      : std_logic;
  signal wrTrigMaskxS                   : std_logic;
  signal wrTrigValxS                    : std_logic;
  signal wrTrigFlagsxS                  : std_logic;
  signal wrTrigDelayxS                  : std_logic;
  signal wrTrigUnitSelxS                : std_logic;
  signal wrDividerxS                    : std_logic;
  signal resetCmdxR                     : std_logic;
  signal flagDemuxxS                    : std_logic;
  signal flagFilterxS                   : std_logic;
  signal flagExtClkxS                   : std_logic;
  signal flagFallingEdgexS              : std_logic;
  signal wrFwdxS, wrBwdxS               : std_logic;
  -- signal armControllerxS                : std_logic;

  signal runBwdCntxSN, runBwdCntxSP               : std_logic;
  signal samplerDataxDN, samplerDataxDP           : std_logic_vector(31 downto 0) := (others => '0');
  signal samplerDataValidxSN, samplerDataValidxSP : std_logic                     := '0';
  signal stopxS                                   : std_logic;
  signal armTriggerxSN, armTriggerxSP             : std_logic                     := '0';
  signal TrigDataInxDN, TrigDataInxDP             : std_logic_vector(31 downto 0) := (others => '0');
  signal TrigDataInValidxSN, TrigDataInValidxSP   : std_logic                     := '0';
  signal TrigDataOutxDN, TrigDataOutxDP           : std_logic_vector(31 downto 0) := (others => '0');
  signal TrigDataOutValidxSN, TrigDataOutValidxSP : std_logic                     := '0';
  signal runxSN, runxSP                           : std_logic                     := '0';
  signal startSamplingxSN, startSamplingxSP       : std_logic;
  signal resetxRN, resetxRP                       : std_logic                     := '1';

  -- the 'gate' controlls the data going out to the sample memory
  signal gatexSN, gatexSP                 : std_logic                     := '0';
  signal gateDataxD                       : std_logic_vector(31 downto 0) := (others => '0');
  signal gateDataValidxS                  : std_logic                     := '0';
  signal dataOutxDN, dataOutxDP           : std_logic_vector(31 downto 0) := (others => '0');
  signal dataOutValidxSN, dataOutValidxSP : std_logic                     := '0';
  signal lastStopxSN, lastStopxSP         : std_logic                     := '0';
  signal samplingDonexS                   : std_logic;
  signal irqxSN, irqxSP                   : std_logic                     := '0';
  -- signal fwdCntDataValidxS : std_logic;
  
begin

  -- add a flip flop on the reset path to improve timing and to resync the
  -- reset signal to the sampling clock. The added delay of one cycle is negligible on this part of the system
  resetxRN <= ResetxRI or resetCmdxR;

  process (ClkxCI)
  begin
    if rising_edge(ClkxCI) then
      resetxRP <= resetxRN;
    end if;
  end process;

  memoryClkxCO  <= ClkxCI;
  flagExtClkxSO <= FlagExtClkxS;
  Inst_decoder : decoder
    port map (
      ClkxCI             => ClkxCI,
      opcodexDI          => opcodexDI,
      dataxDI            => cmdDatxDI,
      executexSI         => executexSI,
      wrTrigUnitSelxSO   => wrTrigUnitSelxS,
      wrTrigMaskxSO      => wrTrigMaskxS,
      wrTrigValxSO       => wrTrigValxS,
      wrTrigFlagsxSO     => wrTrigFlagsxS,
      wrTrigDelayxSO     => wrTrigDelayxS,
      wrSpeedxSO         => wrDividerxS,
      wrFwdxSO           => wrFwdxS,
      wrBwdxSO           => wrBwdxS,
      flagDemuxxSO       => flagDemuxxS,
      flagFilterxSO      => flagFilterxS,
      flagExtClkxSO      => flagExtClkxS,
      flagFallingEdgexSO => flagFallingEdgexS,
      runxSO             => startSamplingxSN,
      resetCorexRO       => resetCmdxR);

  Inst_sync : sync
    generic map (
      CHANNELS => CHANNELS)
    port map (
      ClkxCI      => ClkxCI,
      ClkxCIB     => ClkxCBI,
      ChannelsxDI => channelsxDI,
      FilterxS    => flagFilterxS,
      DemuxxS     => flagDemuxxS,
      FallingxS   => flagFallingEdgexS,
      SyncedxDO   => syncedInputxDN);

  -- insert a register between syncer and sampler to improve timing (allows
  -- longer routing distances and provides registers for retiming)
  process (ClkxCI)
  begin  --
    if ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      syncedInputxDP   <= syncedInputxDN;
      startSamplingxSP <= startSamplingxSN;
    end if;
  end process;

  Inst_sampler : sampler
    generic map (
      CLOCK_DIV_WIDTH => 16)
    port map (
      ClkxCI         => ClkxCI,
      RstxRI         => '0',
      RunxSI         => startSamplingxSP,
      samplexDI      => syncedInputxDP,
      cfgDataxDI     => cmdDatxDI(15 downto 0),
      wrDividerxSI   => wrDividerxS,
      samplexDO      => samplerDataxDN,
      sampleValidxSO => samplerDataValidxSN,
      sampleRatexSO  => open
      );

  sampleClkxSO <= '0';  -- Lukas: I removed this feature due to performance issues

  runBwdCntxSN <= startSamplingxSP;

  -- insert a register between sampler and counter to improve timing (allows
  -- longer routing distances and provides registers for retiming)
  process (ClkxCI)
  begin  --
    if ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      samplerDataxDP <= samplerDataxDN;
      if resetxRP = '1' then
        samplerDataValidxSP <= '0';
        runBwdCntxSP        <= '0';
      else
        samplerDataValidxSP <= samplerDataValidxSN;
        runBwdCntxSP        <= runBwdCntxSN;
      end if;
    end if;
  end process;

  -- delay the start signal for the number of backward samples
  Inst_sampleCounter_bwd : sampleCounter
    generic map (
      CNT_WIDTH => FWD_BWD_WIDTH)
    port map (
      ClkxCI         => ClkxCI,
      RstxRI         => resetxRP,
      samplexDI      => samplerDataxDP,
      sampleValidxSI => samplerDataValidxSP,
      runxSI         => runBwdCntxSP,
      wrTopxSI       => wrBwdxS,
      cfgDataxDI     => cmdDatxDI(FWD_BWD_WIDTH-1 downto 0),
      donexSO        => armTriggerxSN,
      samplexDO      => TrigDataInxDN,
      sampleValidxSO => TrigDataInValidxSN
      );

  -- insert a register between counter and trigger to improve timing (allows
  -- longer routing distances and provides registers for retiming)
  process (ClkxCI)
  begin  --
    if ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      TrigDataInxDP <= TrigDataInxDN;
      if resetxRP = '1' then
        armTriggerxSP      <= '0';
        TrigDataInValidxSP <= '0';
      else
        armTriggerxSP      <= armTriggerxSN;
        TrigDataInValidxSP <= TrigDataInValidxSN;
      end if;
    end if;
  end process;

  Inst_trigger : trigger
    generic map (
      DELAY_LEN => 16,
      UNITS_LEN => TRIG_UNITS_LEN)
    port map (
      ClkxCI           => ClkxCI,
      RstxRI           => resetxRP,
      DatxDI           => TrigDataInxDP,
      DatValidxSI      => TrigDataInValidxSP,
      DemuxedxSI       => flagDemuxxS,
      WriteUnitxSI     => wrTrigUnitSelxS,
      CfgDatxDI        => cmdDatxDI,
      WriteMaskxSI     => wrTrigMaskxS,
      WriteMatchValxSI => wrTrigValxS,
      WriteFlagsxSI    => wrTrigFlagsxS,
      WriteDelayxSI    => wrTrigDelayxS,
      ArmxSI           => armTriggerxSP,
      RunxSO           => runxSN,
      DatxDO           => TrigDataOutxDN,
      DatValidxSO      => TrigDataOutValidxSN
      );

  -- insert a register between counter and trigger to improve timing (allows
  -- longer routing distances and provides registers for retiming)
  process (ClkxCI)
  begin  --
    if ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      TrigDataOutxDP <= TrigDataOutxDN;
      if resetxRP = '1' then
        runxSP              <= '0';
        TrigDataOutValidxSP <= '0';
      else
        runxSP              <= runxSN;
        TrigDataOutValidxSP <= TrigDataOutValidxSN;
      end if;
    end if;
  end process;



  -- delay the fire signal from the trigger for the number of forward samples
  Inst_SampleCounter_fwd : sampleCounter
    generic map (
      CNT_WIDTH => FWD_BWD_WIDTH)
    port map (
      ClkxCI         => ClkxCI,
      RstxRI         => resetxRP,
      samplexDI      => TrigDataOutxDP,
      sampleValidxSI => TrigDataOutValidxSP,
      runxSI         => runxSP,
      wrTopxSI       => wrFwdxS,
      cfgDataxDI     => cmdDatxDI(FWD_BWD_WIDTH-1 downto 0),
      donexSO        => stopxS,
      samplexDO      => gateDataxD,
      sampleValidxSO => gateDataValidxS
      );

  dataOutxDN      <= gateDataxD;
  dataOutValidxSN <= (not stopxS) and gateDataValidxS;

  -- generate an iterrupt request when sampling was done (when stop goes high)
  lastStopxSN    <= stopxS;
  samplingDonexS <= (not lastStopxSP) and stopxS;

  irqxSN <= (irqxSP or samplingDonexS) and (not sample_clrflagxSI);

  process (ClkxCI)
  begin  -- process
    
    if ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      dataOutxDP <= dataOutxDN;
      if resetxRP = '1' then
        dataOutValidxSP <= '0';
        irqxSP          <= '0';
        lastStopxSP     <= '0';
      else
        dataOutValidxSP <= dataOutValidxSN;
        irqxSP          <= irqxSN;
        lastStopxSP     <= lastStopxSN;
      end if;
    end if;
  end process;

  memoryOut            <= dataOutxDP;
  memoryWrite          <= dataOutValidxSP;
  sample_finish_IRQxSO <= irqxSP;

end Behavioral;

