--/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   FPGA Design 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/06/15 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* Namin convention:  http://dz.ee.ethz.ch/en/information/hdl-help/vhdl-naming-conventions.html
--*
--***************************************************************************************************************
--*
--* This entity counts the number of samples once the input signal runxSI goes
--* high. RunxSI has to stay high for the entire counting process. Once the
--* counter has reached configured top value donexSO goes high and stays high until the
--* entity is resetted.
--*
--***************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

entity sampleCounter_tb is
end sampleCounter_tb;

architecture tb of sampleCounter_tb is

  component sampleCounter
    generic (
      CNT_WIDTH : integer);
    port (
      ClkxCI         : in  std_logic;
      RstxRI         : in  std_logic;
      samplexDI      : in  std_logic_vector(31 downto 0);
      sampleValidxSI : in  std_logic;
      runxSI         : in  std_logic;
      wrTopxSI       : in  std_logic;
      cfgDataxDI     : in  std_logic_vector(CNT_WIDTH-1 downto 0);
      donexSO        : out std_logic;
      samplexDO      : out std_logic_vector(31 downto 0);
      sampleValidxSO : out std_logic);
  end component;

  -- component generics
  constant CNT_WIDTH : integer := 16;

  -- component ports
  signal ClkxCI         : std_logic                              := '0';
  signal RstxRI         : std_logic                              := '1';
  signal samplexDI      : std_logic_vector(31 downto 0)          := (others => '0');
  signal sampleValidxSI : std_logic                              := '0';
  signal runxSI         : std_logic                              := '0';
  signal wrTopxSI       : std_logic                              := '0';
  signal cfgDataxDI     : std_logic_vector(CNT_WIDTH-1 downto 0) := (others => '0');
  signal donexSO        : std_logic;
  signal samplexDO      : std_logic_vector(31 downto 0);
  signal sampleValidxSO : std_logic;

  -- clock
  signal Clk : std_logic := '1';

  signal cycleCntxDN, cycleCntxDP : integer := 0;

begin  -- tb

  -- component instantiation
  DUT : sampleCounter
    generic map (
      CNT_WIDTH => CNT_WIDTH)
    port map (
      ClkxCI         => ClkxCI,
      RstxRI         => RstxRI,
      samplexDI      => samplexDI,
      sampleValidxSI => sampleValidxSI,
      runxSI         => runxSI,
      wrTopxSI       => wrTopxSI,
      cfgDataxDI     => cfgDataxDI,
      donexSO        => donexSO,
      samplexDO      => samplexDO,
      sampleValidxSO => sampleValidxSO);

  -- clock generation
  ClkxCI <= not ClkxCI after 5 ns;

  -- waveform generation
  WaveGen_Proc : process
  begin
    -- insert signal assignments here
    wait for 100 ns;
    wait until ClkxCI = '1';
    RstxRI         <= '0';
    wait until rising_edge(ClkxCI);
    cfgDataxDI     <= x"0003";
    wrTopxSI       <= '1';
    sampleValidxSI <= '1';
    wait until rising_edge(ClkxCI);
    runxSI         <= '1';
    wrTopxSI       <= '0';
    wait until rising_edge(ClkxCI) and donexSO = '1';
    runxSI         <= '0';
    RstxRI         <= '1';
    sampleValidxSI <= '0';
    wait until rising_edge(ClkxCI);
    RstxRI         <= '0';
    cfgDataxDI     <= x"0003";
    wrTopxSI       <= '1';
    wait until rising_edge(ClkxCI);
    runxSI         <= '1';
    wrTopxSI       <= '0';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '1';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '0';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '1';
    wait until rising_edge(ClkxCI); 
    sampleValidxSI <= '0';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '1';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '0';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '1';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '0';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '1';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '0';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '1';
    wait until rising_edge(ClkxCI);
    sampleValidxSI <= '0';

    wait;
    
  end process WaveGen_Proc;


  cycleCntxDN <= cycleCntxDP + 1;
  samplexDI <= std_logic_vector(conv_unsigned(cycleCntxDP,32));

  process (ClkxCI)
  begin  -- process
    if ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      cycleCntxDP <= cycleCntxDN;
    end if;
  end process;
  
  

end tb;

