--/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   FPGA Design 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/06/15 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* This is the file based testbench for triggerUnit.vhd
--* It reads stimuli and expected values form a file and uses them to check the
--* model under test (MUT)
--*
--***************************************************************************************************************


-- based on: http://eesun.free.fr/DOC/vhdlref/refguide/language_overview/test_benches/reading_and_writing_files_with_text_i_o.htm
library std;
use std.textio.all;
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;

entity triggerUnitTb is
end triggerUnitTb;

architecture tb of triggerUnitTb is
  
  component triggerUnit
    generic (
      DELAY_LEN : integer);
    port (
      RstxR         : in  std_logic;
      ClkxCI        : in  std_logic;
      MaskxDI       : in  std_logic_vector(31 downto 0);
      MatchValxDI   : in  std_logic_vector(31 downto 0);
      DemuxedxSI    : in  std_logic;
      CfgChannelxDI : in  std_logic_vector(4 downto 0);
      CfgSerialxSI  : in  std_logic;
      CfgNotxSI     : in  std_logic;
      CfgChainxSI   : in  std_logic;
      CfgNextLvlxSI : in  std_logic;
      CfgFirexSI    : in  std_logic;
      CfgDelayxDI   : in  std_logic_vector(DELAY_LEN-1 downto 0);
      DInxDI        : in  std_logic_vector(31 downto 0);
      DInValidxSI   : in  std_logic;
      MatchChainxSI : in  std_logic;
      FiredxSI      : in  std_logic;
      ActivexSI     : in  std_logic;
      DOutxDO       : out std_logic_vector(31 downto 0);
      DOutValidxSO  : out std_logic;
      MatchChainxSO : out std_logic;
      FiredxSO      : out std_logic;
      ActivexSO     : out std_logic);
  end component;

  -- configuration
  constant SIM_DATA_FILE_NAME : string := "../testbenches/triggerUnitTb.sim";  -- file with stimuli and expected results
  constant REPORT_FILE_NAME   : string := "../testbenches/triggerUnitTb.rpt";  -- filename for simulation report

  constant MUT_PIPELINE_LEN : integer := 6;  -- specifies the number of pipeline registers in the MUT

  constant STIM_APPL_TIME : time := 20 ns;  -- time between rising Clk and stim application
  constant RESP_ACQU_TIME : time := 80 ns;  -- time between rising edge and response acquisition
  constant CLK_PERIOD     : time := 100 ns;  -- clock signal period

  -- testbench signals
  signal tbDonexS : std_logic := '0';  -- set to one by the testbench when the simulation is done
  -- signal stimulDataEndxS : std_logic := '0';  -- set by the testbench when stimuli data ends

  -- MUT signals
  signal RstxR         : std_logic                     := '1';
  signal ClkxCI        : std_logic                     := '0';
  signal MaskxDI       : std_logic_vector(31 downto 0) := (others => '0');
  signal MatchValxDI   : std_logic_vector(31 downto 0) := (others => '0');
  signal DemuxedxSI    : std_logic                     := '0';
  signal CfgChannelxDI : std_logic_vector(4 downto 0)  := (others => '0');
  signal CfgSerialxSI  : std_logic                     := '0';
  signal CfgNotxSI     : std_logic                     := '0';
  signal CfgChainxSI   : std_logic                     := '0';
  --signal CfgChainxSI   : std_logic                     := '0';
  signal CfgNextLvlxSI : std_logic                     := '0';
  signal CfgFirexSI    : std_logic                     := '0';
  signal CfgDelayxDI   : std_logic_vector(15 downto 0) := (others => '0');
  signal DInxDI        : std_logic_vector(31 downto 0) := (others => '0');
  signal DInValidxSI   : std_logic                     := '0';
  signal MatchChainxSI : std_logic                     := '0';
  signal FiredxSI      : std_logic                     := '0';
  signal ActivexSI     : std_logic                     := '0';
  signal DOutxDO       : std_logic_vector(31 downto 0);
  signal DOutValidxSO  : std_logic                     := '0';
  signal MatchChainxSO : std_logic                     := '0';
  signal FiredxSO      : std_logic                     := '0';
  signal ActivexSO     : std_logic                     := '0';

  -- delay register to align requested response signals with MUT
  type respDelayDOutType is array (MUT_PIPELINE_LEN-1 downto 0) of std_logic_vector(31 downto 0);
  signal expRespDelayDOutxD       : respDelayDOutType                             := (others => (others => '-'));
  signal expRespDelayDOutValidxD  : std_logic_vector(MUT_PIPELINE_LEN-1 downto 0) := (others => '-');
  signal expRespDelayMatchChainxS : std_logic_vector(MUT_PIPELINE_LEN-1 downto 0) := (others => '-');
  signal expRespDelayFiredxS      : std_logic_vector(MUT_PIPELINE_LEN-1 downto 0) := (others => '-');
  signal expRespDelayActivexS     : std_logic_vector(MUT_PIPELINE_LEN-1 downto 0) := (others => '-');

begin  -- tb

  -- generate a 50% duty cycle clock signal
  clockProcess : process
  begin
    ClkxCI <= '0';
    wait for CLK_PERIOD/2;
    ClkxCI <= '1';
    wait for CLK_PERIOD/2;
  end process;

  simProcess : process

    variable inputLine  : line;         -- use to read data from the stim file
    variable outputLine : line;         -- used to write the sim report
    variable okFlag     : boolean;
    file stimFile       : text is in SIM_DATA_FILE_NAME;  -- stimuli file
    file reportFile     : text is out REPORT_FILE_NAME;   -- report file
    variable errCnt     : integer := 0;  -- used to count the number of failing stimuli

    variable vRstxR         : std_logic                     := '0';
    variable vClkxCI        : std_logic                     := '0';
    variable vMaskxDI       : std_logic_vector(31 downto 0) := (others => '0');
    variable vMatchValxDI   : std_logic_vector(31 downto 0) := (others => '0');
    variable vDemuxedxSI    : std_logic                     := '0';
    variable vCfgChannelxDI : std_logic_vector(4 downto 0)  := (others => '0');
    variable vCfgSerialxSI  : std_logic                     := '0';
    variable vCfgNotxSI     : std_logic                     := '0';
    variable vCfgChainxSI   : std_logic                     := '0';
    --variable vCfgChainxSI   : std_logic                     := '0';
    variable vCfgNextLvlxSI : std_logic                     := '0';
    variable vCfgFirexSI    : std_logic                     := '0';
    variable vCfgDelayxDI   : std_logic_vector(15 downto 0) := (others => '0');
    variable vDInxDI        : std_logic_vector(31 downto 0) := (others => '0');
    variable vDInValidxSI   : std_logic                     := '0';
    variable vMatchChainxSI : std_logic                     := '0';
    variable vFiredxSI      : std_logic                     := '0';
    variable vActivexSI     : std_logic                     := '0';
    variable vDOutxDO       : std_logic_vector(31 downto 0) := (others => '-');
    variable vDOutValidxSO  : std_logic                     := '-';
    variable vMatchChainxSO : std_logic                     := '-';
    variable vFiredxSO      : std_logic                     := '-';
    variable vActivexSO     : std_logic                     := '-';

    variable stimCntxD     : integer := 0;
    type reportDelayType is array (MUT_PIPELINE_LEN downto 0) of line;  -- used to aling the report file lines with the mut output
    variable reportDelayxS : reportDelayType;

-- a is the expected value, b is the actual MUT response
    function checkResp (a : std_logic; b : std_logic)
      return boolean is
    begin
      if (a = '-') or (a = b) then  -- if expected value is '-' we don't check
        return true;
      else
        return false;
      end if;
    end checkResp;

    function checkRespVect (a : std_logic_vector(31 downto 0); b : std_logic_vector(31 downto 0))
      return boolean is
      variable okFlag : boolean := true;  -- if one bit fails the check this goes to false
    begin
      for i in 0 to 31 loop
        okFlag := okFlag and checkResp(a(i), b(i));
      end loop;  -- i
      return okFlag;
    end checkRespVect;

    -- purpose: compares actual output and expected values and creates the report
    procedure checkOutput (
      variable reportDelayxS : in    reportDelayType;
      variable errCnt        : inout integer;  -- counts the number of failing stimuli
      file reportFile        :       text) is  -- file handle for report data
      variable okFlag     : boolean := true;
      variable outputLine : line;
    begin  -- checkOutput
      -- check MUT output values
      okFlag := checkRespVect(expRespDelayDOutxD(MUT_PIPELINE_LEN-1), DOutxDO);
      okFlag := okFlag and checkResp(expRespDelayDOutValidxD(MUT_PIPELINE_LEN-1), DOutValidxSO);
      okFlag := okFlag and checkResp(expRespDelayMatchChainxS(MUT_PIPELINE_LEN-1), MatchChainxSO);
      okFlag := okFlag and checkResp(expRespDelayFiredxS(MUT_PIPELINE_LEN-1), FiredxSO);
      okFlag := okFlag and checkResp(expRespDelayActivexS(MUT_PIPELINE_LEN-1), ActivexSO);

      outputLine := reportDelayxS(MUT_PIPELINE_LEN);
      -- outputLine := inputLine;

      if not okFlag then
        -- the MUT did not give the expected result, report the error
        errCnt := errCnt + 1;
        write(outputLine, " | ");
        -- append the MUT result to the report
        hwrite(outputLine, DOutxDO, right, 9);
        write(outputLine, DOutValidxSO, right, 2);
        write(outputLine, MatchChainxSO, right, 2);
        write(outputLine, FiredxSO, right, 2);
        write(outputLine, ActivexSO, right, 2);
        report "MUT did not match expected result, see report file" severity warning;
      end if;

      write(outputLine, "     at " & time'image(now));
      --if outputLine /= null then
      if outputLine'length > 0 then
        writeline(reportFile, outputLine);
      end if;
      --end if;
    end checkOutput;
    
  begin  -- sim process

    
    
    loop
      -- stop simulation at end of file
      if endfile(stimFile) then
        assert false report "stimuli file finished" severity note;
        exit;
      end if;

      -- get one line from the file
      readline(stimFile, inputLine);

      next when inputLine'length = 0;   -- ignore empty lines
      if inputLine(1) = '%' then        -- ignore comments starting with %
        if stimCntxD = 0 then
          -- this is the comment header of the stimuli file, copy it to the report
          writeline(reportFile, inputLine);
        end if;
        next;
      end if;

      -- wait for rising clk edge (equals time 0 within one clock cycle)
      wait until ClkxCI'event and ClkxCI = '1';

      -- copy the stimuli line to the report shift register
      if reportDelayxS(MUT_PIPELINE_LEN) /= null then
        deallocate(reportDelayxS(MUT_PIPELINE_LEN));
      end if;
      for i in MUT_PIPELINE_LEN downto 1 loop
        reportDelayxS(i) := reportDelayxS(i-1);
      end loop;  -- i
      reportDelayxS(0) := new string'(inputLine(inputLine'low to inputLine'high));

      stimCntxD := stimCntxD + 1;

      -- shift the expected response registers synchronously with the MUT
      expRespDelayDOutxD       <= expRespDelayDOutxD(MUT_PIPELINE_LEN-2 downto 0) & vDoutxDO;
      expRespDelayDOutValidxD  <= expRespDelayDOutValidxD(MUT_PIPELINE_LEN-2 downto 0) & vDOutValidxSO;
      expRespDelayMatchChainxS <= expRespDelayMatchChainxS(MUT_PIPELINE_LEN-2 downto 0) & vMatchChainxSO;
      expRespDelayFiredxS      <= expRespDelayFiredxS(MUT_PIPELINE_LEN-2 downto 0) & vFiredxSO;
      expRespDelayActivexS     <= expRespDelayActivexS(MUT_PIPELINE_LEN-2 downto 0) & vActivexSO;

      -- read the stimuli
      read(inputLine, vRstxR, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      hread(inputLine, vMaskxDI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      hread(inputLine, vMatchValxDI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vDemuxedxSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vCfgChannelxDI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vCfgSerialxSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vCfgNotxSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vCfgChainxSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vCfgNextLvlxSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vCfgFirexSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      -- wait for 1 ns;
      hread(inputLine, vCfgDelayxDI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      hread(inputLine, vDInxDI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vDInValidxSI, okFlag);
      -- wait for 1 ns;
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vMatchChainxSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vFiredxSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vActivexSI, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      -- read the expected results
      hread(inputLine, vDOutxDO, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vDOutValidxSO, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vMatchChainxSO, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vFiredxSO, okFlag);
      assert okFlag report "error reading stimuli file" severity error;
      read(inputLine, vActivexSO, okFlag);
      assert okFlag report "error reading stimuli file" severity error;

      -- Note: hread does not support reading - as hex value. Therefore we
      -- replace the data signal with - if data-valid signal is '-'
      if vDOutValidxSO = '-' then
        vDOutxDO := "------------------------------";
        report "replacing data with x""--------""" severity note;
      end if;

      -- wait for stimuli application time
      wait for STIM_APPL_TIME;

      -- apply the stimuli
      RstxR         <= vRstxR;
      MaskxDI       <= vMaskxDI;
      MatchValxDI   <= vMatchValxDI;
      DemuxedxSI    <= vDemuxedxSI;
      CfgChannelxDI <= vCfgChannelxDI;
      CfgSerialxSI  <= vCfgSerialxSI;
      CfgNotxSI     <= vCfgNotxSI;
      CfgChainxSI   <= vCfgChainxSI;
      CfgNextLvlxSI <= vCfgNextLvlxSI;
      CfgFirexSI    <= vCfgFirexSI;
      CfgDelayxDI   <= vCfgDelayxDI;
      DInxDI        <= vDInxDI;
      DInValidxSI   <= vDInValidxSI;
      MatchChainxSI <= vMatchChainxSI;
      FiredxSI      <= vFiredxSI;
      ActivexSI     <= vActivexSI;

      wait for (RESP_ACQU_TIME-STIM_APPL_TIME);

      -- only report if the pipeline is filled with valid stimuli data
      if stimCntxD >= MUT_PIPELINE_LEN then

        checkOutput(reportDelayxS, errCnt, reportFile);

      end if;
    end loop;

    -- check the last results which are still in the pipeline
    for i in 0 to MUT_PIPELINE_LEN-1 loop
      -- wait for rising clk edge (equals time 0 within one clock cycle)
      wait until ClkxCI'event and ClkxCI = '1';

      --reportDelayxS(MUT_PIPELINE_LEN-1 downto 1) := reportDelayxS(MUT_PIPELINE_LEN-2 downto 0);
      -- copy the stimuli line to the report shift register
      if reportDelayxS(MUT_PIPELINE_LEN) /= null then
        deallocate(reportDelayxS(MUT_PIPELINE_LEN));
      end if;
      for i in MUT_PIPELINE_LEN downto 1 loop
        reportDelayxS(i) := reportDelayxS(i-1);
      end loop;  -- i
      reportDelayxS(0) := new string'("<no stimuli data>");

      stimCntxD := stimCntxD + 1;

      -- shift the expected response registers synchronously with the MUT
      expRespDelayDOutxD(MUT_PIPELINE_LEN-1 downto 0) <= expRespDelayDOutxD(MUT_PIPELINE_LEN-2 downto 0) & vDoutxDO;
      expRespDelayDOutValidxD                         <= expRespDelayDOutValidxD(MUT_PIPELINE_LEN-2 downto 0) & vDOutValidxSO;
      expRespDelayMatchChainxS                        <= expRespDelayMatchChainxS(MUT_PIPELINE_LEN-2 downto 0) & vMatchChainxSO;
      expRespDelayFiredxS                             <= expRespDelayFiredxS(MUT_PIPELINE_LEN-2 downto 0) & vFiredxSO;
      expRespDelayActivexS                            <= expRespDelayActivexS(MUT_PIPELINE_LEN-2 downto 0) & vActivexSO;

      wait for STIM_APPL_TIME;

      -- apply default values, we don't have any more stimuli
      RstxR         <= '0';
      MaskxDI       <= x"00000000";
      MatchValxDI   <= x"00000000";
      DemuxedxSI    <= '0';
      CfgChannelxDI <= "00000";
      CfgSerialxSI  <= '0';
      CfgNotxSI     <= '0';
      CfgChainxSI   <= '0';
      CfgNextLvlxSI <= '0';
      CfgFirexSI    <= '0';
      CfgDelayxDI   <= x"0000";
      DInxDI        <= x"00000000";
      DInValidxSI   <= '0';
      MatchChainxSI <= '0';
      FiredxSI      <= '0';
      ActivexSI     <= '0';

      wait for (RESP_ACQU_TIME-STIM_APPL_TIME);

      checkOutput(reportDelayxS, errCnt, reportFile);
      
    end loop;

    tbDonexS <= '1';                    -- tb is done
    wait until ClkxCI'event and ClkxCI = '1';

    if errCnt > 0 then
      
      outputLine := new string'("The MUT failed in ");
      write(outputLine, errCnt);
      write(outputLine, string'(" test cases"));
    else
      outputLine := new string'("All tests passed");
      
    end if;
    writeline(reportFile, outputLine);

    wait;
    
  end process;


  -- the model under test
  MUT : triggerUnit
    generic map (
      DELAY_LEN => 16)
    port map (
      RstxR         => RstxR,
      ClkxCI        => ClkxCI,
      MaskxDI       => MaskxDI,
      MatchValxDI   => MatchValxDI,
      DemuxedxSI    => DemuxedxSI,
      CfgChannelxDI => CfgChannelxDI,
      CfgSerialxSI  => CfgSerialxSI,
      CfgNotxSI     => CfgNotxSI,
      CfgChainxSI   => CfgChainxSI,
      CfgNextLvlxSI => CfgNextLvlxSI,
      CfgFirexSI    => CfgFirexSI,
      CfgDelayxDI   => CfgDelayxDI,
      DInxDI        => DInxDI,
      DInValidxSI   => DInValidxSI,
      MatchChainxSI => MatchChainxSI,
      FiredxSI      => FiredxSI,
      ActivexSI     => ActivexSI,
      DOutxDO       => DOutxDO,
      DOutValidxSO  => DOutValidxSO,
      MatchChainxSO => MatchChainxSO,
      FiredxSO      => FiredxSO,
      ActivexSO     => ActivexSO);


end tb;
