/**
* copyright 2011 Lukas Schrittwieser, all rights reserved
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

/*
* 
*/

// max. counter compare for a max dutycycle of approx 0.4
#define D_MAX	150


// define the desired converter output voltage (mulitply value in Volts with ~50) 
#define V_DESIRED	((int16_t)622)

#define P_GAIN		16
#define T_INTEGRAL	1

// saturation value for accumulator (calc this to fulfill the max D value)
#define ACCU_TOP	(uint16_t)(((uint16_t)D_MAX*T_INTEGRAL)/P_GAIN)

// saturation value for accu + err
#define ACCU_PLUS_ERR_TOP	(uint32_t)(((uint32_t)D_MAX)/P_GAIN)

volatile uint16_t sysTick = 0;
volatile uint8_t powerUpDone = 0;


void init (void);

int main (void)
{
	DDRB |= (1<<PB1); //Speaker-Output
	PORTB |= (1<<PB1); //Make a Beep
	_delay_ms(25);
	PORTB &= ~(1<<PB1);
	
	init();
	
	uint8_t ovpcnt = 0;
	uint8_t buzzerstate = 0;
	
	while(1)
	{
		if(buzzerstate == 0) //buzzer is off
		{
			if(PINB & (1<<PB3)) //overvoltage!
			{
				ovpcnt++;
				if(ovpcnt >= 250)
				{
					PORTB |= (1<<PB1); //buzzer on
					buzzerstate = 1;
					ovpcnt = 0;
				}
			}
			else
			{
				ovpcnt = 0;
			}
					
		}
		else if(buzzerstate == 1) //buzzer is on
		{
			_delay_ms(10);
			if(!(PINB & (1<<PB3))) //no more overvoltage!
			{
				ovpcnt++;
				if(ovpcnt >= 15)
				{
					PORTB &= ~(1<<PB1); //buzzer on
					buzzerstate = 0;
					ovpcnt = 0;
				}
			}
			
		}
		
	}
}


ISR(TIMER0_COMPA_vect)
{
	ADCSRA |= (1<<ADSC);	// start adc
}


void init (void)
{
	// start pll and wait for lock
	PLLCSR = /*(1<<LSM) | */ (1<<PLLE); // select low speed mode
	while (!(PLLCSR&(1<<PLOCK)))
		asm("nop");

	// timer 1
	PLLCSR |= (1<<PCKE);	// select pll clock for timer 1
	OCR1B = 0x00;			// set duty cycle to 0
	OCR1C = 0xff;			// full pwm resolution 
	GTCCR = (1<<PWM1B) | (1<<COM1B1);	
	TCCR1 = (0<<CS12) | (1<<CS10);		// select div by 1 prescaler
	

	// adc
	ADMUX =  (1<<REFS2) | (1<<REFS1) | (0<<REFS0) | (0<<MUX1) | (1<<MUX0);	// select 2.56V reference and ADC1
	ADCSRA = (1<<ADEN) | (1<<ADIE) | (1<<ADPS2) | (1<<ADPS1);	// select div by 64 prescaler and Int enable flag
	DIDR0 = (1<<ADC1D);	// disable digital inputs

	PORTB &= ~(1<<4);	
	DDRB = (1<<4);	// mosfet output

	DDRB |= (1<<PB1); //Speaker output
	
	DDRB &= ~(1<<PB3); //Overvoltage-FET-Input

	// timer 0
	OCR0A = ((8000000/8)/16000)-1;
	TCCR0A = (1<<WGM01);	// select CTC mode
	TCCR0B = (1<<CS01);		// select div by 8 prescaler
	TIMSK = (1<<OCIE0A);


	sei();

}


// ADC conversion is complete -> calculate loop controller and update duty cycle
ISR(ADC_vect)
{
	static int16_t accu = 0;
	int16_t err;
	int32_t dsqr_s;
	uint16_t dsqr_u;

	//sei();

	// get ADC sample 
	
	// calculate error signal 
	err = V_DESIRED - ((int16_t)ADC);
	
	//OCR1A = (uint8_t)((ADC/4));

	// implement the accumulator
	accu = accu + err;
	if (accu < 0)
		accu = 0;
	if (accu > ACCU_TOP)
		accu = ACCU_TOP;

	//OCR1A = (uint8_t)((accu&0xff));

	dsqr_s = (accu/T_INTEGRAL) + err;
	// limit dsqr to the allowed values
	if (dsqr_s < 0)
		dsqr_s = 0;
	if (dsqr_s > ACCU_PLUS_ERR_TOP)
		dsqr_s = ACCU_PLUS_ERR_TOP;

	// apply proportional gain
	dsqr_u = ((uint16_t)dsqr_s) * P_GAIN;

	// calc dutycycle by taking the square root of d square
	OCR1B = (dsqr_u&0xff);
}
