/*
 *  Copyright (C) 2006:Michael Poppitz 2011: Mario Mauerer, Lukas Schrittwieser
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 *
 */
package org.sump.analyzer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.Timer;

import org.sump.util.Properties;


// TODO: when the dialog is closed using the window decoration's close function, close() is not called

/**
 * GUI Component that allows the user to control the device and start captures.
 * <p>
 * Its modelled after JFileChooser and should allow for non-dialog implementations
 * making it somewhat reusable.
 * 
 * @version 0.7
 * @author Michael "Mr. Sump" Poppitz
 * 
 * @version 2.0
 * @author Mario Mauerer
 * 
 *
 */
public class DeviceController extends JComponent implements ActionListener, Runnable, Configurable {
	/** dialog showing and waiting for user action */
	public final static int IDLE = 0;
	/** capture currently running */
	public final static int RUNNING = 1;
	/** capture / dialog aborted by user */
	public final static int ABORTED = 2;
	/** capture finished */
	public final static int DONE = 3;
	
	//Trigger Config Flags:
	//Serial-Flag:
	public final static byte SERIALFLAG = 8;
	//Not-Flag:
	public final static byte NOTFLAG = 9;
	//Chain-Flag
	public final static byte CHAINFLAG = 10;
	//Next-Level-Flag:
	public final static byte NLFLAG = 11;
	//Fire-Flag:
	public final static byte FIREFLAG = 12;
	
	
	/**
	 * Creates an array of check boxes, adds it to the device controller and returns it.
	 * @param label label to use on device controller component
	 * @return array of created check boxes
	 */
 	private JCheckBox[] createChannelList(JPanel pane, GridBagConstraints constraints) {
		JCheckBox[] boxes = new JCheckBox[32];

		Container container = new Container();
		container.setLayout(new GridLayout(1, 32));
	
		for (int col = 31; col >= 0; col--) {
			JCheckBox box = new JCheckBox();
			box.setEnabled(false);
			container.add(box);
			if ((col % 8) == 0 && col > 0)
				container.add(new JLabel());
			boxes[col] = box;
		}
	
		pane.add(container, constraints);
		return (boxes);
	}
 	
 	private JCheckBox[] createMatrixList(int len, JPanel pane, GridBagConstraints constraints) {
		JCheckBox[] boxes = new JCheckBox[len];

		Container container = new Container();
		container.setLayout(new GridLayout(1, len));
	
		for (int col = 0; col<len; col++) {
			JCheckBox box = new JCheckBox();
			box.setEnabled(false);
			container.add(box);
			boxes[col] = box;
		}
		for(int col = len; col<triggerStages; col++)
		{
			container.add(new JLabel(""));
		}
	
		pane.add(container, constraints);
		return (boxes);
	}
	
	private static GridBagConstraints createConstraints(int x, int y, int w, int h, double wx, double wy) {
	 	GridBagConstraints gbc = new GridBagConstraints();
		gbc.fill = GridBagConstraints.BOTH;
	        gbc.insets = new Insets(2, 2, 2, 2);
	 	gbc.gridx = x; gbc.gridy = y;
		gbc.gridwidth = w; gbc.gridheight = h;
	        gbc.weightx = wx; gbc.weighty = wy;
	        return (gbc);
	}
	
	/**
	 * Constructs device controller component.
	 *
	 */
	public DeviceController() 
	{
		super();
		setLayout(new GridBagLayout());
		setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

		device = new Device();

		
		// connection pane
		JPanel connectionPane = new JPanel();
		connectionPane.setLayout(new GridLayout(2, 2, 5, 5));
		connectionPane.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder("Connection Settings"),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)
		));		
		
	

		String[] ports = Device.getPorts();
		//NOT USED FOR DEBUGGING BITHOUNDLESS - uncomment the following if no BitHound is present but the GUI should run.
		
		if (ports.length==0)
		{
			// Now device was found in the network -> abort software
			JOptionPane.showMessageDialog(this,
					"Error:\n\n"
					+ "No device was found in the network.\n\n"
					+ "Make sure the device is:\n"
					+ " - connected to the network\n"
					+ " - turned on and properly programmed\n"
					+ " - has got an IP address\n",
					"Communication Error",
					JOptionPane.ERROR_MESSAGE
				);
			System.exit(-1); 
		}
	
		portSelect = new JComboBox(ports);
		connectionPane.add(new JLabel("Analyzer Port:"));
		connectionPane.add(portSelect);
		
		add(connectionPane, createConstraints(0, 0, 1, 1, 0.0, 0.0));
		
		
// settings pane
		JPanel settingsPane = new JPanel();
		settingsPane.setLayout(new GridLayout(11, 2,5, 5));
		settingsPane.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder("General Settings"),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)
		));
		
		//Select the Trigger Sources:
		String[] sources = {
			"Internal"
		//	"External / Rising",
		//	"External / Falling"
		};
		sourceSelect = new JComboBox(sources);
		sourceSelect.addActionListener(this);
		settingsPane.add(new JLabel("Sampling Clock:"));
		settingsPane.add(sourceSelect);
		
		//Select the Sampling-Frequencies:
		String[] speeds = {
			"400MHz", "200MHz", "100MHz", "50MHz", "20MHz", "10MHz", "5MHz", "2MHz", "1MHz",
			"500kHz", "200kHz", "100kHz", "50kHz", "20kHz", "10kHz"
		};
		speedSelect = new JComboBox(speeds);
		speedSelect.setSelectedIndex(1);
		speedSelect.addActionListener(this);
		settingsPane.add(new JLabel("Sampling Rate:"));
		settingsPane.add(speedSelect);

		//Select the amount of samples to take:
		String[] sizes = {"32M", "16M", "8M", "4M", "2M", "1M", "512K",
			"256K", "128K", "64K", "32K", "16K", "8K", "4K",
			"2K", "1K", "512", "256", "128", "64"
		};
		sizeSelect = new JComboBox(sizes);
		sizeSelect.setSelectedIndex(7);
		sizeSelect.addActionListener(this);
		settingsPane.add(new JLabel("Recording Size:"));
		settingsPane.add(sizeSelect);

		//Display the resulting sampling time:	
		settingsPane.add(new JLabel("==> Sampling-Time:"));
		samplingtime = new JLabel(""); //create the samplingtime-JLabel
		settingsPane.add(samplingtime); //add it to the panel
		updateSamplingtime(); //display the current sampling-time.
		
		//Select the channel groups to be displayed:
		Container groups = new Container();
		groups.setLayout(new GridLayout(2, 2));
		channelGroup = new JCheckBox[4];
		for (int i = 0; i < (channelGroup.length); i++) 
		{
			channelGroup[i] = new JCheckBox(Integer.toString(i));
			channelGroup[i].setSelected(true);
			groups.add(channelGroup[i]);
		}
		
		settingsPane.add(new JLabel("Display Channel Groups     "));
		settingsPane.add(groups);
		

		//Enable the Filter:
		filterEnable = new JCheckBox("Enable");
		filterEnable.setSelected(true);
		filterEnable.setEnabled(false);
		settingsPane.add(new JLabel("Noise Filter: "));
		settingsPane.add(filterEnable);
		
		
		add(settingsPane, createConstraints(0, 1, 1, 1, 0.0, 0.0));

		
//Configure the Trigger: Trigger-Pane:
		JPanel triggerPane = new JPanel();
		triggerPane.setLayout(new GridBagLayout());
		triggerPane.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder("Trigger Settings"),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)
		));
		
		//Trigger-Mode-Selection:
		String[] trigmode = {"Enabled", "Free-Running"};
		triggerEnable = new JComboBox(trigmode);
		triggerEnable.addActionListener(this);
		triggerEnable.setSelectedIndex(1);
		triggerPane.add(new JLabel("Trigger Mode:"), createConstraints(0, 0, 1, 1, 0.0, 1.0));
		triggerPane.add(triggerEnable, createConstraints(1, 0, 1, 1, 0.5, 1.0));
		
		
		//Before/After-Ratio: Only Visible, when Trigger enabled!
		String[] ratios = {"0/100", "25/75", "50/50", "75/25", "100/0"};
		ratioSelect = new JComboBox(ratios);
		ratioSelect.setSelectedIndex(2);
		triggerPane.add(new JLabel("Before/After Ratio: "), createConstraints(0, 1, 1, 1, 0.5, 1.0));
		triggerPane.add(ratioSelect, createConstraints(1, 1, 1, 1, 0.5, 1.0));
		
		//Trigger Type: (Simple/Complex) (Only visible, when Trigger enabled!)
		String[] types = {"Simple", "Advanced"};
		triggerTypeSelect = new JComboBox(types);
		triggerTypeSelect.addActionListener(this);
		triggerPane.add(new JLabel("Type: "), createConstraints(0, 2, 1, 1, 0.5, 1.0));
		triggerPane.add(triggerTypeSelect, createConstraints(1, 2, 1, 1, 0.5, 1.0));
		
		
		//Simple Trigger Options: Only visible, whenn Trigger enabled + simple mode!
		String[] simpleop = {"Rising Edge", "Falling Edge"};
		simpleTrigMode = new JComboBox(simpleop);
		triggerPane.add(new JLabel("Trigger on:"), createConstraints(3,2,1,1,0.5,1.0));
		triggerPane.add(simpleTrigMode,createConstraints(4,2,1,1,0.5,1.0));
		
		//Simple Trigger Channel Selection:
		String[] simpleCh = {"0","1","2","3","4","5","6","7","8","9","10","11","12","13","14","15","16","17","18","19","20","21","22","23","24","25","26","27","28","29","30","31"};
		simpleTrigCh = new JComboBox(simpleCh);
		triggerPane.add(new JLabel("on Channel:"), createConstraints(5,2,1,1,0.5,1.0));
		triggerPane.add(simpleTrigCh,createConstraints(6,2,1,1,0.5,1.0));
		
		triggerPane.add(new JLabel(" "),createConstraints(0,3,1,1,1.0,1.0));
		
		
		triggerPane.add(new JLabel("<html><h3>Advanced Trigger Settings:</h3></html>"),createConstraints(0,4,1,2,1.0,1.0)); 
		
		//Advanced-Trigger-Configuration
		triggerStages = device.getTriggerStageCount();
		
		//Stage-Selection/Activation:
		stageSel = new Container();
		stageSel.setLayout(new GridLayout(1, triggerStages));
		stageSelGroup= new JCheckBox[triggerStages];
		
		for (int i = 0; i < triggerStages; i++) {
			stageSelGroup[i] = new JCheckBox(Integer.toString(i));
			stageSelGroup[i].setSelected(false);
			stageSelGroup[i].addActionListener(this);
			stageSel.add(stageSelGroup[i]);
		}
		triggerPane.add(new JLabel("Activate Stages:"),createConstraints(0,6,1,1,1.0,1.0));
		triggerPane.add(stageSel,createConstraints(1,6,1,1,1.0,1.0));
		
		//Stage-Panes:
		triggerStageTabs = new JTabbedPane();
		triggerMask = new JCheckBox[triggerStages][];
		triggerValue = new JCheckBox[triggerStages][];
		triggerDelay = new JTextField[triggerStages];
		triggerMode = new JComboBox[triggerStages];
		triggerChannel = new JComboBox[triggerStages];
		stageNotFlag = new JCheckBox[triggerStages];

		for (int i = 0; i < triggerStages; i++) {
			JPanel stagePane = new JPanel();
			stagePane.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
			stagePane.setLayout(new GridBagLayout());

			String[] modes = {"Parallel", "Serial"};
			triggerMode[i] = new JComboBox(modes);
			stagePane.add(new JLabel("Mode:", JLabel.RIGHT), createConstraints(2, 0, 1, 1, 0.5, 1.0));
			stagePane.add(triggerMode[i], createConstraints(3, 0, 1, 1, 0.5, 1.0));
			triggerMode[i].addActionListener(this);
			
			String[] channels = {
				"0",  "1",  "2",  "3",  "4",  "5",  "6",  "7",  "8",  "9",  "10", "11", "12", "13", "14", "15",
				"16", "17", "18", "19", "20", "21", "22", "23", "24", "25", "26", "27", "28", "29", "30", "31"
			};
			triggerChannel[i] = new JComboBox(channels);
			stagePane.add(new JLabel("Channel:", JLabel.RIGHT), createConstraints(4, 0, 1, 1, 0.5, 1.0));
			stagePane.add(triggerChannel[i], createConstraints(5, 0, 1, 1, 0.5, 1.0));
			
			stagePane.add(new JLabel("31"), createConstraints(1, 1, 1, 1, 1.0, 1.0));
			stagePane.add(new JLabel("0", JLabel.RIGHT), createConstraints(5, 1, 1, 1, 1.0, 1.0));
			stagePane.add(new JLabel("Mask:"), createConstraints(0, 2, 1, 1, 1.0, 1.0));
			triggerMask[i] = createChannelList(stagePane, createConstraints(1, 2, 5, 1, 1.0, 1.0));
			stagePane.add(new JLabel("Value:"), createConstraints(0, 3, 1, 1, 1.0, 1.0));
			triggerValue[i] = createChannelList(stagePane, createConstraints(1, 3, 5, 1, 1.0, 1.0));

			
			stagePane.add(new JLabel("Delay:", JLabel.RIGHT), createConstraints(4, 4, 1, 1, 0.5, 1.0));
			triggerDelay[i] = new JTextField("0");
			stagePane.add(triggerDelay[i], createConstraints(5, 4, 1, 1, 0.5, 1.0));
			triggerStageTabs.add("Stage " + i, stagePane);
			
			stagePane.add(new JLabel("Invert the trigger-output of this stage:"),createConstraints(2,4,1,1,1.0,0.0));
			stageNotFlag[i] = new JCheckBox();
			stagePane.add(stageNotFlag[i],createConstraints(3,4,1,1,1.0,0.0));
		}
		triggerPane.add(triggerStageTabs, createConstraints(0, 7, 7, 1, 1.0, 1.0));
				
				
		//Add the Stage-Interconnection-Matrix:
		stageConfig = new JPanel();
		stageConfig.setLayout(new GridBagLayout());
		stageConfig.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder("Stage Interconnection"),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)
		));
		
		stageMatrix = new JCheckBox[triggerStages][]; //The AND-Matrix
		stageOrBoxes = new JCheckBox[triggerStages]; //OR-Selection-Blocks
		stageNlBoxes = new JCheckBox[triggerStages]; //Next-Level-Selection-Blocks

		
		stageConfig.add(new JLabel("Stage"),createConstraints(0,0,1,1,0.0,0.0));
		for(int i = 0; i<triggerStages; i++)
		{
			stageConfig.add(new JLabel(Integer.toString(i).concat(" ==>"),JLabel.RIGHT),createConstraints(0,i+1,1,1,0.0,0.0));
			stageMatrix[i] = createMatrixList(i+1,stageConfig,createConstraints(1,i+1,1,1,0.0,0.0));
		}
		//Get the Background color and safe it: (needed later for Next-Level-Coloring)
		background = new Color[1];
		background[0] = stageMatrix[0][0].getBackground();
		
		//Now, add the "AND"s:
		Container andcontainer = new Container();
		andcontainer.setLayout(new GridLayout(1, triggerStages));
		for (int col = triggerStages-1; col >= 0; col--) {
			JLabel andbox = new JLabel(" \136");
			andcontainer.add(andbox);
		}
		stageConfig.add(andcontainer, createConstraints(1,9,1,1,0.0,0.0));
		
		
		//The OR-Blocks:
		stageOrBoxes = createMatrixList(triggerStages,stageConfig,createConstraints(1,10,1,1,0.0,0.0));
		stageNlBoxes= createMatrixList(triggerStages-1,stageConfig,createConstraints(1,11,1,1,0.0,0.0));
	
		stageConfig.add(new JLabel("==> OR ==> Start Sampling"),createConstraints(9,10,1,1,0.0,0.0));
		stageConfig.add(new JLabel("==> Create next armed stage"),createConstraints(9,11,1,1,0.0,0.0));

		//Add an Action Listeners for every Matrix-checkbox:
		for(int i = 0; i<triggerStages; i++)
		{
			for(int j = 0;j<=i;j++)
			{
				stageMatrix[i][j].addActionListener(this);
			}
			stageOrBoxes[i].addActionListener(this);
			

		}
		for(int i = 0;i<triggerStages-1;i++)
		{
			stageNlBoxes[i].addActionListener(this);
		}
		
		//Display the Interconnection-String:
		InterConnLabel = new JLabel("");
		stageConfig.add(new JLabel("Selected Stage-Interconnection:"),createConstraints(10,0,1,1,0.0,0.0));
		stageConfig.add(InterConnLabel,createConstraints(10,1,1,triggerStages*2,0.0,0.0));
		
		
		triggerPane.add(stageConfig,createConstraints(0,8,7,1,1.0,1.0));
		
		
		add(triggerPane, createConstraints(1, 0, 2, 2, 1.0, 0.5));//Don't Delete this!
		
	
		// progress pane
		JPanel progressPane = new JPanel();
		progressPane.setLayout(new BorderLayout());
		progressPane.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder("Progress"),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)
		));
		progress = new JProgressBar(0, 100);
		progressPane.add(progress, BorderLayout.CENTER);
		add(progressPane, createConstraints(0, 2, 3, 1, 1.0, 0));

		add(new JLabel(), createConstraints(0, 3, 1, 1, 0.5, 0));
		
		captureButton = new JButton("Capture");
		captureButton.addActionListener(this);
		add(captureButton, createConstraints(1, 3, 1, 1, 0.5, 0));

		JButton cancel = new JButton("Close");
		cancel.addActionListener(this);
		add(cancel, createConstraints(2, 3, 1, 1, 0.5, 0));
		
		capturedData = null;
		timer = null;
		worker = null;
		status = IDLE;
	}

	/**
	 * Internal method that initializes a dialog and add this component to it.
	 * @param frame owner of the dialog
	 */
	private void initDialog(JFrame frame) {
		// check if dialog exists with different owner and dispose if so
		if (dialog != null && dialog.getOwner() != frame) {
			dialog.dispose();
			dialog = null;
		}
		// if no valid dialog exists, create one
		if (dialog == null) {
			dialog = new JDialog(frame, "Capture", true);
			dialog.getContentPane().add(this);
			dialog.setResizable(false);
			dialog.setSize(this.getPreferredSize());
			// dialog.pack();
		}
		// reset progress bar
		progress.setValue(0);

		// sync dialog status with device
		updateFields();
	}

	/**
	 * Return the device data of the last successful run.
	 * 
	 * @return device data
	 */
	public CapturedData getDeviceData() {
		return (capturedData);
	}

	/**
	 * Extracts integers from strings regardless of trailing trash.
	 * 
	 * @param s string to be parsed
	 * @return integer value, 0 if parsing fails
	 */
	private int smartParseInt(String s) {
		int val = 0;
	
		try {
			for (int i = 1; i <= s.length(); i++)
				val = Integer.parseInt(s.substring(0, i));
		} catch (NumberFormatException E) {}
		
		return (val);
	}

	
	/**
	 * Sets the enabled state of all configuration components of the dialog.
	 * @param enable <code>true</code> to enable components, <code>false</code> to disable them
	 */
	private void setDialogEnabled(boolean enable) {
		triggerEnable.setEnabled(enable);
		captureButton.setEnabled(enable);
		portSelect.setEnabled(enable);
		speedSelect.setEnabled(enable);
		sizeSelect.setEnabled(enable);
		updateFields(enable);
	}


	/**
	 * This function updates the trigger-Section in the GUI:
	 * @param triggersimple says wheter the trigger is in simple mode or not
	 * @param triggerenable says wheter the trigger is enabled or not.
	 */
	private void updateTriggerGUI(boolean triggersimple, boolean triggerenable)
	{
		
		triggerStages = device.getTriggerStageCount(); //Get the number of Trigger-Stages from the Device
			
		
		//Display (or don't) the advanced section:
		//Display the Stage-Selection-Checkboxes:
		//Only display the next checkbox, if the previous is selected:
		stageSelGroup[0].setEnabled((!triggersimple) && triggerenable); //The first one always needs to be enabled...
		if(!stageSelGroup[0].isEnabled())
		{
			stageSelGroup[0].setEnabled(false);
			stageSelGroup[0].setSelected(false);
		}
		for(int i = 0; i<triggerStages-1; i++)
		{
			if(stageSelGroup[i].isSelected())
				stageSelGroup[i+1].setEnabled(true); //enable the next one 
			else
			{
				stageSelGroup[i+1].setEnabled(false); //otherwise, disable and de-select the following ones. 
				stageSelGroup[i+1].setSelected(false);
			}
		}
		
		//Display the Trigger-Options: (The Trigger-Tabs)
		//Display the stages that have been enabled with StageSelGroup.
		int channels = device.getAvailableChannelCount();
		for (int stage = 0; stage < triggerStages; stage++) 
		{
			for (int i = 0; i < channels; i++) {
				triggerMask[stage][i].setEnabled(stageSelGroup[stage].isSelected());
				triggerValue[stage][i].setEnabled(stageSelGroup[stage].isSelected());
			}
			for (int i = channels; i < 32; i++) {
				triggerMask[stage][i].setEnabled(false);
				triggerValue[stage][i].setEnabled(false);
			}
			triggerStageTabs.setEnabledAt(stage, (stageSelGroup[stage].isSelected()));
			triggerDelay[stage].setEnabled((stageSelGroup[stage].isSelected()));
			triggerMode[stage].setEnabled(stageSelGroup[stage].isSelected());
			
			if (triggerenable && triggerMode[stage].getSelectedIndex() == 1) {
				triggerChannel[stage].setEnabled(true);
			} else {
				triggerChannel[stage].setEnabled(false);
			}
			
			stageNotFlag[stage].setEnabled(stageSelGroup[stage].isSelected());
			
		}
			
		//Display the Trigger-Interconnection-Matrix:
		//Always display the first row (if stage0 is selected..):
		if(stageSelGroup[0].isSelected())
			stageMatrix[0][0].setEnabled(true);
		else
		{
			stageMatrix[0][0].setEnabled(false);
			stageMatrix[0][0].setSelected(false);
		}
		
		//Only display a following row/col, if the previous row/col is used:
		//first, disable all boxes to create a defined environment:
		for(int i = 1;i<triggerStages;i++)
		{
			for(int j = 0;j<=i;j++)
			{
				stageMatrix[i][j].setEnabled(false);
			}
		}
		//Now, enable the boxes that are allowed:
		for(int i = 1; i<triggerStages; i++)
		{
			for(int m=1;m<=i;m++) 
			{
				if(stageMatrix[i-1][m-1].isSelected() && stageSelGroup[i].isSelected()) //the upper left box is selected!
				{
					stageMatrix[i][m].setEnabled(true); //select the lower right box as the most right boundary
					//Now, enable the box left from the just enabled one as well:
					stageMatrix[i][m-1].setEnabled(true);
					m = i+1; //exit the loop, no more boxes need to be selected.
				}
			}
		}
		
		//If a cross has been made in a row: deactivate the other boxes in this row:
		for(int i = 0; i<triggerStages; i++)
		{
			for(int m=0; m<=i; m++)
			{
				if(stageMatrix[i][m].isSelected()) //there is a cross in this row!
				{
					for(int k = 0; k<=i; k++)
					{
						if(k!=m)
							stageMatrix[i][k].setEnabled(false); //disable the other boxes in this row
					}
					m= i+1; //exit the for-loop, everything has been done. 
				}
			}
		}
		
		//now, look for the row with a disabled, but set box. Disable and de-set this and all folowing ones! ==> Cleanup!
		int breakoutvar = 0; //needed to jump out of the for-loops. 
		for(int i = 1;i<triggerStages;i++)
		{
			for(int j = 0;j<=i;j++)
			{
				if( stageMatrix[i][j].isSelected() && !(stageMatrix[i][j].isEnabled())) //if there is a disabled but selected box ==> bad! ==> cleanup needed.
				{
					//Clean up ALL boxes BELOW the one that has just been discovered to be faulty!
					for(int k = i;k<triggerStages;k++)
					{
						for(int l = 0;l<=k;l++)
						{
							stageMatrix[k][l].setEnabled(false);
							stageMatrix[k][l].setSelected(false);
						}
					}
					breakoutvar = 1;//Cleanup finished, leave the for-loops.
					break;
				}
				if(breakoutvar == 1)
					break;
			}
			if(breakoutvar == 1)
				break;
		}
		
		//Display the OR-Boxes:
		//If there is a cross in a column, select the OR-Box and enable the corresponding Nl-Box.
		int checkboxsetflag = 0; //flag to check if there is a selected box somewhere.
		int stagecnt = 0; //determine, how many stages are active - used later for coloring!
		for(int i =0; i<triggerStages; i++)
		{
			checkboxsetflag = 0;
			for(int j = 0;j<triggerStages-i;j++) 
			{
				if(stageMatrix[j+i][i].isSelected())//is there a cross in the column?
				{
					checkboxsetflag = 1;
				}
			}
			//There has been a cross in this column
			if(checkboxsetflag == 1) //there is a cross in the matrix!
			{
				if(i<triggerStages-1) //needed because there are only triggerStages-1 Next-Level-Boxes
					stageNlBoxes[i].setEnabled(true);
				stageOrBoxes[i].setSelected(true);
				stagecnt++;
			}
			else //no cross ==> No Or-Box has to be selected.
			{
				if(i<triggerStages-1)//there are only triggerStages-1 of the Nl-Checkboxes...
				{
					stageNlBoxes[i].setEnabled(false);
					stageNlBoxes[i].setSelected(false);
				}
				stageOrBoxes[i].setSelected(false);
			}
			
		}
		
		//find the last Or-Box and disable it's Nl-Box: (a Next-Level in the Last Block makes no sense)
		for(int i = 0;i<triggerStages-1;i++)
		{
			if(!(stageOrBoxes[i+1].isSelected()))
			{
				if(i<triggerStages-1)
				{
					stageNlBoxes[i].setEnabled(false);
					stageNlBoxes[i].setSelected(false);
				}
			}
		}
		
		//In the Following, the Level-Blocks (if selected) are colored:
		//Create the Colors: 
		Stagecolor = new Color[8];
		Stagecolor[0] = Color.blue;
		Stagecolor[1] = Color.green;
		Stagecolor[2] = Color.red;
		Stagecolor[3] = Color.gray;
		Stagecolor[4] = Color.orange;
		Stagecolor[5] = Color.magenta;
		Stagecolor[6] = Color.pink; 
		Stagecolor[7] = Color.cyan; 
		
		int start = 0;
		int colorcnt = 0;
		int colorflag = 0;
		for(int i = 0;i<stagecnt;i++)
		{
			if(i<triggerStages-1)
			{
				if(stageNlBoxes[i].isSelected()) //if we have a next-level-block selected:
				{
					for(int col = start; col<=i;col++)
					{
						for(int row = triggerStages-1;row>=col;row--)
						{	
							stageMatrix[row][col].setBackground(Stagecolor[colorcnt]); //color the blocks up to our selected block
						}
					}
					start = i+1; //color the next block from the next column
					colorcnt++; //choose next color
					colorflag = 1; //we have something colored
				}
			}
			if(colorflag == 1) //color the entire rest!
			{
				for(int col = start; col<=i;col++)
				{
					for(int row = triggerStages-1;row>=col;row--)
					{	
						stageMatrix[row][col].setBackground(Stagecolor[colorcnt]);
					}
				}
			}
				
		}
		//Now: Remove the areas that may not be colored:
		for(int i = 0;i<triggerStages;i++)
		{
			if(!(stageOrBoxes[i].isSelected()))//if there is no Or-Flag in a Col: Make that col white!
			{
				for(int row = triggerStages-1;row>=i;row--)
				{	
					stageMatrix[row][i].setBackground(background[0]); //set the old, normal background value!
				}
			}
		}
		//If no Nl-Flag at all: Color nothing!
		int whiteflag = 0;
		for(int i = 0;i<triggerStages-1;i++)
		{
			if((stageNlBoxes[i].isSelected()))//if there is still a flag...
				whiteflag = 1;//color nothing
		}
		if(whiteflag == 0)//all colours can be removed!
		{
			for(int col = start; col<triggerStages;col++)
			{
				for(int row = triggerStages-1;row>=col;row--)
				{	
					stageMatrix[row][col].setBackground(background[0]); //set the old, normal background value!
				}
			}
		}
		
		//Now, display the interconnection-String:
		updateTriggerInterconnString(stagecnt,triggersimple);
	
		
	}
	
	/**
	 * This function displays a string representing the trigger-interconnection
	 */
	private void updateTriggerInterconnString(int stagecnt, boolean triggersimple)
	{
		String InterConnString;
		InterConnString = new String("<html>");
		
		int hascross = 0;
		int NLset = 0;
		int lengthcnt = 0;
		
		for(int col = 0; col<stagecnt; col++)
		{
			hascross = 0;
			lengthcnt = 0;
			for(int row = col; row<triggerStages; row++)
			{
				if(stageMatrix[row][col].isSelected())
				{
					if(col >= 1)
					{
						if(stageOrBoxes[col-1].isSelected() && hascross == 0 && NLset == 0)
						{
							InterConnString = InterConnString.concat(" OR ");
							lengthcnt+=3;
						}	
					}
					
					if(hascross == 0)
						InterConnString = InterConnString.concat("(");
					
					InterConnString = InterConnString.concat("Stage").concat(Integer.toString(row));
					lengthcnt+=6;
					NLset = 0;
					hascross = 1;
				}
				
				if(row < triggerStages-1)
				{
					if(stageMatrix[row+1][col].isSelected() && hascross == 1)
					{
						InterConnString = InterConnString.concat(" AND ");
						lengthcnt+=3;
					}
					else if(hascross == 1)
					{
						InterConnString = InterConnString.concat(")");
						row = triggerStages + 1;
						if(lengthcnt>=22)
						{
							InterConnString = InterConnString.concat("<br>");
							lengthcnt = 0;
						}
					}
				}
				if(row == triggerStages-1 && hascross == 1)
				{
					InterConnString = InterConnString.concat(")");	
					
					if(lengthcnt>=22)
					{
						InterConnString = InterConnString.concat("<br>");
						lengthcnt = 0;
					}	
				}

				
			}
			
			if(col < triggerStages-1)
			{
				if(stageNlBoxes[col].isSelected())
				{
					InterConnString = InterConnString.concat("<br>==>><br>");
					NLset = 1;
					lengthcnt = 0;
				}
			}
	
		}
		
		if(InterConnString.length() < 7)
		{
			InterConnString = "<html>Nothing selected.<br> Please select the <br>desired triggerstage-<br>interconnection " +
					"<br>in the matrix on the left.<br> Only applicable in<br>Advanced Mode.</html>";
		}

		InterConnLabel.setText(InterConnString);
	}
	
	
	/** activates / deactivates dialog options according to device status */
	public void updateFields() {
		updateFields(true);
	}
	
/**
 * This Function updates the entire Trigger-GUI.
 */
	/** activates / deactivates dialog options according to device status */
	private void updateFields(boolean enable) {

		boolean triggerenable = device.isTriggerEnabled(); //Check, if Trigger is really enabled
		
		//Check, if the trigger is in simple mode:
		boolean triggersimple; 
		String value;
		 value = (String)triggerTypeSelect.getSelectedItem(); //Check, if Simple or Advanced
		if(value.equals("Simple"))
			triggersimple = true; //Trigger in Simple mode
		else if(value.equals("Advanced"))
			triggersimple = false; //Trigger in Advanced mode
		else
			triggersimple = false; //Default Assignment...
		
		
		simpleTrigCh.setEnabled(triggersimple && triggerenable); //This displays the Trigger channels when simple and enabled.
		simpleTrigMode.setEnabled(triggersimple && triggerenable); //This displays the simple trigger option when the trigger is in simple mode.
		triggerTypeSelect.setEnabled(triggerenable); //This displays or hides the Trigger-Type-Selection-Dialog depending of the Trigger's enabled-status.
		ratioSelect.setEnabled(triggerenable); //This displays or hides the ratio-select-field depending on the trigger-enable-status.
		
		updateTriggerGUI(triggersimple, triggerenable); //Display the Trigger-Options in the GUI
	
		//Enable the Filters if applicable:
		filterEnable.setEnabled(device.isFilterAvailable() && enable);
		//Enable the Channel-Groups that are possible:
		for (int i = 0; i < channelGroup.length; i++)
			channelGroup[i].setEnabled(enable && (i < device.getAvailableChannelCount() / 8));
		
		//Enable the Speed-option if it is applicable:
		speedSelect.setEnabled(device.getClockSource() == Device.CLOCK_INTERNAL);
		
		updateSamplingtime(); //update the displayed sampling-time!
	}
	
	
	/**
	 * This function calculates the (time-)length of the current sampling-block
	 */
	//this function updates the displayed sampling-time. 
	private void updateSamplingtime() 
	{
		long speed;
		long size;
		float period;
		float time;
		String value;
		
		//Check, if the time can be displayed at all...:
		value = (String)sourceSelect.getSelectedItem();
		if (!(value.equals("Internal"))) //If NOT Internally Triggered
		{
			samplingtime.setText("Not applicable"); //Sampling-Freq. not known...
			return;
		}
		
		//Get the Speed
		value = (String)speedSelect.getSelectedItem();
		speed = smartParseInt(value);
		if (value.indexOf("M") > 0)
			speed *= 1000000;
		else if (value.indexOf("k") > 0)
			speed *= 1000;
		
		//Get the Number of Samples:
		value = (String)sizeSelect.getSelectedItem();
		size = smartParseInt(value);
		if (value.indexOf("K") > 0)
			size *= 1024;
		if (value.indexOf("M") > 0)
			size *= 1024*1024;
		
		//Calculate the period and the sampling-time:
		period = 1/(float)speed;
		time = period * (float)size;
		
		//Adjust the values to us, ms, s, min:
		if(time < (float)0.001)
		{
			time = time * (float)1000000;
			samplingtime.setText((Float.toString(time)).concat(" us"));
		}
		else if(time < (float)1)
		{
			time = time * (float)1000;
			samplingtime.setText((Float.toString(time)).concat(" ms"));
		}
		else if(time > (float)60)
		{
			time = time / (float)60;
			samplingtime.setText((Float.toString(time)).concat(" min"));
		}
		else
		{
			samplingtime.setText((Float.toString(time)).concat(" s"));
		}
		
	}
	
	/**
	 * This Function updates the BitHound-Device, it sets all Parameters according to the GUI.
	 */
	
	/** writes the dialog settings to the device */
	private void updateDevice() {
		String value;

		// set clock source
		value = (String)sourceSelect.getSelectedItem();
		if (value.equals("Internal")) {
			device.setClockSource(Device.CLOCK_INTERNAL);
		} else {
			if (value.equals("External / Rising"))
				device.setClockSource(Device.CLOCK_EXTERNAL_RISING);
			else 
				device.setClockSource(Device.CLOCK_EXTERNAL_FALLING);
		}

		// set sample rate
		value = (String)speedSelect.getSelectedItem();
		int f = smartParseInt(value);
		if (value.indexOf("M") > 0)
			f *= 1000000;
		else if (value.indexOf("k") > 0)
			f *= 1000;
		try{
		device.setRate(f);
		}
		catch(IOException e)
		{
			JOptionPane.showMessageDialog(this,
					"Error:\n\n"
					+ "With this hardware, this sampling-frequency\n"
					+ "is not applicable.\n"
					+ "Please select another sampling-frequency!\n",
					e.getMessage(),
					JOptionPane.ERROR_MESSAGE);
		}
		
		// set sample count
		value = (String)sizeSelect.getSelectedItem();
		int s = smartParseInt(value);
		if (value.indexOf("K") > 0)
			s *= 1024;
		if (value.indexOf("M") > 0)
			s *= 1024*1024;
		device.setSize(s);
		
		// set before / after ratio
		value = (String)ratioSelect.getSelectedItem();
		double r = 0.5;
		if (value.equals("100/0")) r = 0;
		else if (value.equals("25/75")) r = 0.75;
		else if (value.equals("50/50")) r = 0.5;
		else if (value.equals("75/25")) r = 0.25;
		else if (value.equals("0/100")) r = 1;
		device.setRatio(r);
		
		// set filter
		device.setFilterEnabled(filterEnable.isSelected());

	// set trigger
		//Get the stage-Count:
		triggerStages = device.getTriggerStageCount();
		
		//Enable/Disable trigger:
		boolean triggerEnabled;
		value = (String)triggerEnable.getSelectedItem(); //Read the GUI
		if(value.equals("Enabled"))
			triggerEnabled = true;
		else if(value.equals("Free-Running"))
			triggerEnabled = false;
		else
			triggerEnabled = false;
		device.setTriggerEnabled(triggerEnabled); //enable/disable trigger.
		
		//Configure the Trigger:
		if(triggerEnabled) //if the Trigger is switched on - config it!
		{
			//check if simple or advanced mode:
			boolean triggersimple; 
			 value = (String)triggerTypeSelect.getSelectedItem(); //Check, if Simple or Advanced from GUI
			if(value.equals("Simple"))
				triggersimple = true;
			else if(value.equals("Advanced"))
				triggersimple = false;
			else
				triggersimple = false;
			
			if(triggersimple) //if the trigger is in SIMPLE mode
			{
				//Read which mode:
				byte simpletrigmode;
				value = (String)simpleTrigMode.getSelectedItem(); //Check, what mode from Simple mode.
				if(value.equals("Rising Edge"))
					simpletrigmode = 1;
				else if(value.equals("Falling Edge"))
					simpletrigmode = 2;
				else
					simpletrigmode = 0;
				
				//Get the Channels:
				byte simpleCh = 0;
				value = (String)simpleTrigCh.getSelectedItem(); //Get the selected Channel from the GUI
				simpleCh = (byte)smartParseInt(value);
				
				if(simpletrigmode == 1) //Trigger on a rising edge!
				{
					
					//Only set the first trigger-Stage to trigger on a rising edge:
					int trigconfig = 0x00;
					trigconfig |= (1<<SERIALFLAG); //Trigger in Serial Mode
					trigconfig |= (1<<FIREFLAG); //Set the Fire-Flag
					trigconfig |= simpleCh; //Set the desired Channel
					device.setTriggerStage(0,trigconfig,0x00000003,0x01,0x00);
					//Now, config the other Trigger-Stages:
					for(int i = 1; i<triggerStages;i++)
					{
						device.setTriggerStage(i, 0x00, 0x00, 0x00, 0x00); //set them all to zero.
					}
				}
				else if(simpletrigmode == 2)//Trigger on a falling edge!
				{
					//Only set the first trigger-Stage to trigger on a rising edge:
					int trigconfig = 0x00;
					trigconfig |= (1<<SERIALFLAG); //Trigger in Serial Mode
					trigconfig |= (1<<FIREFLAG);
					trigconfig |= simpleCh;
					device.setTriggerStage(0,trigconfig,0x00000003,0x02,0x00);
					//Now, config the other Trigger-Stages:
					for(int i = 1; i<triggerStages;i++)
					{
						device.setTriggerStage(i, 0x00, 0x00, 0x00, 0x00); //set them all to zero.
					}
				}
				
			}
			else //ADVANCED mode ==> Config the Trigger for advanced!
			{
				configAdvancedDevice(); //Configure the Advanced Trigger!
			}
		}
		else//else: trigger is NOT switched on: config it for free-running mode: only set Fire-Flag.
		{
			int trigconfig = 0x00;
			for(int i = 0; i<triggerStages-1;i++)
			{
				device.setTriggerStage(i, trigconfig, 0x00, 0x00, 0x00); //set them all to zero.
			}
			trigconfig |= (1<<FIREFLAG); 
				device.setTriggerStage(triggerStages-1,trigconfig,0x00,0x00,0x00); //Set the fire-Flag in the last Trigger stage only.
		}
	
		
		// set enabled channel groups
		int enabledChannels = 0;
		for (int i = 0; i < channelGroup.length; i++)
			if (channelGroup[i].isSelected())
				enabledChannels |= 0xff << (8 * i);
		device.setEnabledChannels(enabledChannels);
	}

	/**
	 * This function configures the trigger of the device if it is in advanced mode. 
	 */
	private void configAdvancedDevice()
	{
		int[] stagemask;
		int[] stagevalue;
		int[] stageconfig;
		int[] stagedelay;
		
		stagemask = new int[triggerStages];
		stagevalue = new int[triggerStages];
		stageconfig = new int[triggerStages];
		stagedelay = new int[triggerStages];
		
		//Reset the values:
		for(int stage = 0;stage<triggerStages; stage++)
		{
			stagemask[stage] = 0x00;
			stagevalue[stage] = 0x00;
			stageconfig[stage] = 0x00;
			stagedelay[stage] = 0x00;
		}
		
		//find out, how many stages are active: 
		int activestages = 0;
		for(int stages = 0; stages < triggerStages; stages++)
		{
			if(stageSelGroup[stages].isSelected())
				activestages++;		
		}
		
		//first, configure all the Flags and the simple Stuff:
		for(int stage = 0; stage < activestages; stage++) //configure each stage of the trigger with the simple flags/values etc::
		{
			if(stageNotFlag[stage].isSelected())//if the Not-Flag is set
				stageconfig[stage] |= (1<<NOTFLAG);
			
			//Now, check if parallel or serial mode:
			String value;
			 value = (String)triggerMode[stage].getSelectedItem(); //Check, if Parallel or Serial
			if(value.equals("Serial"))
			{
				stageconfig[stage] |= (1<<SERIALFLAG); //Set the Config to Serial mode!
				int channel = triggerChannel[stage].getSelectedIndex(); //Read, which channel has to be triggered on.
				stageconfig[stage] |= channel; //set the cannel.
			}
			
			//set the delay:
			stagedelay[stage] = smartParseInt(triggerDelay[stage].getText());
			
			//set the mask and the value:
			for (int i = 0; i < 32; i++) 
			{
				if (triggerMask[stage][i].isSelected())
					stagemask[stage] |= (1<<i);
				if (triggerValue[stage][i].isSelected())
					stagevalue[stage] |= (1<<i);				
			}
		}
		
	
		//Now, translate the interconnection-matrix into  the fire, chain, and nextlevel-flag:
		int colhascross = 0;
		for(int col = 0;col<activestages; col++)
		{
			colhascross = 0;
			
			for(int row = col;row<triggerStages;row++)
			{
				if(stageMatrix[row][col].isSelected()) //if there is a cross
				{
					if(colhascross == 0)
						colhascross = 1;
					else if(colhascross != 0) // set the flags (setting the flags in unit is useless...)
					{
						stageconfig[row] |= (1<<CHAINFLAG); //set the Chain(AND)-Flag
					}
					if(colhascross == 1 && row == triggerStages-1) //if we are at the bottom, set the fire-flag in any case.
					{
						stageconfig[row] |= (1<<FIREFLAG);
					}
				}
				else //no more crosses - set the fire-Flag  of the last unit that had a cross and check for NextLevel-Flag.
				{
					if(colhascross == 1)
					{
						stageconfig[row-1] |= (1<<FIREFLAG);
						
						//Check for the Next-level-Flag:
						if(col < triggerStages-1)
						{
							if(stageNlBoxes[col].isSelected())//If fireflag set
							{
								stageconfig[row] |= (1<<NLFLAG);
							}
						}
						row = triggerStages; //Exit this loop, no more to do, go for next column. 
					}
				}
			}
		}
		
		//Transfer the configuration to the device!
		System.out.println("Configuration:");
		for(int stage = 0;stage<triggerStages; stage++)
		{
			System.out.println("ADVANCED: Stage:");
			System.out.println(Integer.toString(stage));
			System.out.println("Config:");
			if((stageconfig[stage] & (1<<NOTFLAG)) != 0)
				System.out.print("_NF_");
			if((stageconfig[stage] & (1<<CHAINFLAG)) != 0)
				System.out.print("_CF_");
			if((stageconfig[stage] & (1<<FIREFLAG)) != 0)
				System.out.print("_FF_");
			if((stageconfig[stage] & (1<<NLFLAG)) != 0)
				System.out.print("_NLF_");
			System.out.println(" ");
			
			
			System.out.println("Mask:");
			System.out.println(Integer.toString(stagemask[stage]));
			System.out.println("Value:");
			System.out.println(Integer.toString(stagevalue[stage]));
			System.out.println("Delay:");
			System.out.println(Integer.toString(stagedelay[stage]));
			
			
			device.setTriggerStage(stage, stageconfig[stage], stagemask[stage], stagevalue[stage], stagedelay[stage]);//Transfer the configs to the device!
		}
		
	}
	
	
	/**
	 * Starts capturing from device. Should not be called externally.
	 */
	public void run()  {
		// TODO: need to check if attach was successful
		device.attach(
			(String)portSelect.getSelectedItem(),
			/*smartParseInt((String)portRateSelect.getSelectedItem()*/ 0
		);
	
		status = RUNNING;
		
		try {
			System.out.println(""+System.currentTimeMillis()+" Run started");
			errorMessage = "";
			
			capturedData = device.run();
			System.out.println(""+System.currentTimeMillis()+" Run completed");
			status = DONE;
		} catch (Exception ex) {
			// TODO: could make sense to also return half read captures if array length is corrected
			capturedData = null;
			status = ABORTED;
			System.out.println("Run aborted");
			//if (!(ex instanceof InterruptedException)) {
				errorMessage = ex.getMessage();
				ex.printStackTrace(System.out);
				System.out.println(errorMessage);
			//}
		}
		device.detach();
	}
	
	/**
	 * Properly closes the dialog.
	 * This method makes sure timer and worker thread are stopped before the dialog is closed.
	 *
	 */
	private void close() {
		if (timer != null) {
			timer.stop();
			timer = null;
		}
		if (worker != null) {
			device.stop(); // lets hope no one gets here before device.run() is called
			System.out.println("interrupting worker thread");
			worker.interrupt();
			worker = null;
		}
		dialog.hide();
	}
	
	/**
	 * Starts the capture thread.
	 */
	private void startCapture() {
		try {
			setDialogEnabled(false);
			timer = new Timer(100, this);
			worker = new Thread(this);
			timer.start();
			worker.start();
		} catch(Exception E) {
			E.printStackTrace(System.out);
		}
	}
	
	/**
	 * Handles all action events for this component.
	 */ 
	public void actionPerformed(ActionEvent event) {
		Object o = event.getSource();
		String l = event.getActionCommand();
		
		// ignore all events when dialog is not displayed
		if (dialog == null || !dialog.isVisible())
			return;
		
		if (o == timer) {
			if (status == DONE) {
				System.out.println("DONE -> closing");
				close();
			} else if (status == ABORTED) {
				timer.stop();
				JOptionPane.showMessageDialog(this,
					"Error while trying to communicate with device:\n\n"
					+ "\"" + errorMessage + "\"\n\n"
					+ "Make sure the device is:\n"
					+ " - connected to the specified port\n"
					+ " - turned on and properly programmed\n"
					+ " - set to the selected transfer rate\n",
					"Communication Error",
					JOptionPane.ERROR_MESSAGE
				);
				setDialogEnabled(true);
			} else {
				if(device.isRunning())
					progress.setValue(device.getPercentage());
			}
		} else {
		
			if (o == triggerEnable) {
				updateDevice();
				updateFields();

			} else if (o == sourceSelect) {
				updateDevice();
				updateFields();

			} else if (o == speedSelect) {
				updateDevice();
				updateFields();
				
			} else if (o == sizeSelect){
				updateFields();
			
			} else if (o == stageSelGroup){
				updateFields();
			
			} else if(o == stageMatrix){
				updateFields();
			
			} else if(o==stageOrBoxes){
				updateFields();
				
			} else if(o==stageNlBoxes){
				updateFields();
			
			} else if (l.equals("Capture")) {
				updateDevice();
				startCapture();

			} else if (l.equals("Close")) {
				close();

			} else {
				updateFields();
			}
		}
	}
	
	private void selectByValue(JComboBox box, String value) {
		if (value != null)
			for (int i = 0; i < box.getItemCount(); i++)
				if (value.equals((String)box.getItemAt(i)))
					box.setSelectedIndex(i);
	}

	public void readProperties(Properties properties) {
		selectByValue(portSelect, properties.getProperty(NAME + ".port"));
		selectByValue(sourceSelect, properties.getProperty(NAME + ".source"));
		selectByValue(speedSelect, properties.getProperty(NAME + ".speed"));
		selectByValue(sizeSelect, properties.getProperty(NAME + ".size"));
		selectByValue(ratioSelect, properties.getProperty(NAME + ".ratio"));
		filterEnable.setSelected("true".equals(properties.getProperty(NAME + ".filter")));
		selectByValue(triggerEnable, properties.getProperty(NAME + ".trigmode"));
		selectByValue(triggerTypeSelect, properties.getProperty(NAME + ".triggerType"));
		selectByValue(simpleTrigMode, properties.getProperty(NAME + ".simpletrig"));
		selectByValue(simpleTrigCh, properties.getProperty(NAME + ".simplCh"));
		
		
		for (int stage = 0; stage < triggerStages; stage++) {
			triggerDelay[stage].setText(properties.getProperty(NAME + ".triggerStage" + stage + "Delay"));
			selectByValue(triggerMode[stage], properties.getProperty(NAME + ".triggerStage" + stage + "Mode"));
			selectByValue(triggerChannel[stage], properties.getProperty(NAME + ".triggerStage" + stage + "Channel"));
			
			String mask = properties.getProperty(NAME + ".triggerStage" + stage + "Mask");
			if (mask != null)
				for (int i = 0; i < 32 && i < mask.length(); i++)
					triggerMask[stage][i].setSelected(mask.charAt(i) == '1');
			
			String mask2 = properties.getProperty(NAME + ".stamatrix" + stage + "Matrix");
			if (mask2 != null)
				for (int i = 0; i < 32 && i < mask2.length(); i++)
					stageMatrix[stage][i].setSelected(mask2.charAt(i) == '1');
	
			String value = properties.getProperty(NAME + ".triggerStage" + stage + "Value");
			if (value != null)
				for (int i = 0; i < 32 && i < value.length(); i++)
					triggerValue[stage][i].setSelected(value.charAt(i) == '1');

		}
		
		String group = properties.getProperty(NAME + ".channelGroup");
		if (group != null)
			for (int i = 0; i < 4 && i < group.length(); i++)
				channelGroup[i].setSelected(group.charAt(i) == '1');

		triggerStages = device.getTriggerStageCount();
		
		String stageselection = properties.getProperty(NAME + ".stageSelGroup");
		if (stageselection != null)
			for (int i = 0; i < triggerStages && i < stageselection.length(); i++)
				stageSelGroup[i].setSelected(stageselection.charAt(i) == '1');
		
		String Nlselection = properties.getProperty(NAME + ".stageNlBoxes");
		if (Nlselection != null)
			for (int i = 0; i < triggerStages && i < Nlselection.length(); i++)
				stageNlBoxes[i].setSelected(Nlselection.charAt(i) == '1');
		
		String stageors = properties.getProperty(NAME + ".stageOrBoxes");
		if (stageors != null)
			for (int i = 0; i < triggerStages && i < stageors.length(); i++)
				stageOrBoxes[i].setSelected(stageors.charAt(i) == '1');
		
		updateDevice();
		updateFields();
}
	
	public void writeProperties(Properties properties) {
		properties.setProperty(NAME + ".port", (String)portSelect.getSelectedItem());
		properties.setProperty(NAME + ".source", (String)sourceSelect.getSelectedItem());
		properties.setProperty(NAME + ".speed", (String)speedSelect.getSelectedItem());
		properties.setProperty(NAME + ".size", (String)sizeSelect.getSelectedItem());
		properties.setProperty(NAME + ".ratio", (String)ratioSelect.getSelectedItem());
		properties.setProperty(NAME + ".filter", filterEnable.isSelected()?"true":"false");
		properties.setProperty(NAME + ".trigmode", (String)triggerEnable.getSelectedItem());
		properties.setProperty(NAME + ".triggerType", (String)triggerTypeSelect.getSelectedItem());
		properties.setProperty(NAME + ".simpletrig",(String)simpleTrigMode.getSelectedItem());
		properties.setProperty(NAME + ".simplCh",(String)simpleTrigCh.getSelectedItem());
		
		for (int stage = 0; stage < triggerStages; stage++) {
			properties.setProperty(NAME + ".triggerStage" + stage + "Delay", triggerDelay[stage].getText());
			properties.setProperty(NAME + ".triggerStage" + stage + "Mode", (String)triggerMode[stage].getSelectedItem());
			properties.setProperty(NAME + ".triggerStage" + stage + "Channel", (String)triggerChannel[stage].getSelectedItem());
			
			StringBuffer mask = new StringBuffer();
			for (int i = 0; i < 32; i++)
				mask.append(triggerMask[stage][i].isSelected()?"1":"0");
			properties.setProperty(NAME + ".triggerStage" + stage + "Mask", mask.toString());
	
			StringBuffer value = new StringBuffer();
			for (int i = 0; i < 32; i++)
				value.append(triggerValue[stage][i].isSelected()?"1":"0");
			properties.setProperty(NAME + ".triggerStage" + stage + "Value", value.toString());

			StringBuffer matrix = new StringBuffer();
			for (int i = 0; i < 32; i++)
				matrix.append(stageMatrix[stage][i].isSelected()?"1":"0");
			properties.setProperty(NAME + ".stamatrix" + stage + "Matrix", matrix.toString());
			
		}
		
		StringBuffer group = new StringBuffer();
		for (int i = 0; i < 4; i++)
			group.append(channelGroup[i].isSelected()?"1":"0");
		properties.setProperty(NAME + ".channelGroup", group.toString());
		
		triggerStages = device.getTriggerStageCount();
		StringBuffer stageselection = new StringBuffer();
		for (int i = 0; i < triggerStages; i++)
			stageselection.append(stageSelGroup[i].isSelected()?"1":"0");
		properties.setProperty(NAME + ".stageSelGroup", stageselection.toString());
		
		StringBuffer Nlselection = new StringBuffer();
		for (int i = 0; i < triggerStages; i++)
			Nlselection.append(stageNlBoxes[i].isSelected()?"1":"0");
		properties.setProperty(NAME + ".stageNlBoxes", Nlselection.toString());
		
		
		StringBuffer stageors = new StringBuffer();
		for (int i = 0; i < triggerStages; i++)
			stageors.append(stageOrBoxes[i].isSelected()?"1":"0");
		properties.setProperty(NAME + ".stagOrBoxes", stageors.toString());
		
	}

	
	/**
	 * Displays the device controller dialog with enabled configuration portion and waits for user input.
	 * 
	 * @param frame parent frame of this dialog
	 * @return status, which is either <code>ABORTED</code> or <code>DONE</code>
	 * @throws Exception
	 */
	public int showCaptureDialog(JFrame frame) throws Exception {
		status = IDLE;
		initDialog(frame);
		setDialogEnabled(true);
		dialog.setVisible(true);
		return status;
	}

	/**
	 * Displays the device controller dialog with disabled configuration, starting capture immediately.
	 * 
	 * @param frame parent frame of this dialog
	 * @return status, which is either <code>ABORTED</code> or <code>DONE</code>
	 * @throws Exception
	 */
	public int showCaptureProgress(JFrame frame) throws Exception {
		status = IDLE;
		initDialog(frame);
		startCapture();
		dialog.setVisible(true);
		return status;
	}

	private Thread worker;
	private Timer timer;
	
	private JComboBox portSelect;

	private JComboBox sourceSelect;
	private JComboBox speedSelect;
	private JComboBox sizeSelect;
	private JLabel samplingtime; //used to display the sampling-time.
	private JComboBox ratioSelect;
	private JCheckBox filterEnable;
	private JComboBox triggerEnable;
	private JComboBox triggerTypeSelect;
	private JComboBox simpleTrigMode;
	private JComboBox simpleTrigCh;
	private JTabbedPane triggerStageTabs;
	private JTextField[] triggerDelay;
	private JComboBox[] triggerMode;
	private JCheckBox[][] stageMatrix;
	private JCheckBox[] stageOrBoxes;
	private JCheckBox[] stageNlBoxes;
	private JComboBox[] triggerChannel;
	private JCheckBox[][] triggerMask;
	private JCheckBox[][] triggerValue;
	private JCheckBox[] channelGroup;
	private JCheckBox[] stageSelGroup;
	private JProgressBar progress;
	private JButton captureButton;
	private Container stageSel;
	private JPanel stageConfig;
	private Color[] Stagecolor;
	private Color[] background;
	private JCheckBox[] stageNotFlag;
	private JLabel InterConnLabel;


	
	private JDialog dialog;
	private Device device;
	private CapturedData capturedData;
	
	private int triggerStages;

	
	private int status;
	private String errorMessage;
	
	private static final long serialVersionUID = 1L;
	private static final String NAME = "DeviceController";
	

}
